/*
 * ip6sic - IPv6 Stack Integrity Checker
 *
 * $Id: ip6sic.c,v 1.2 2003/10/01 18:58:15 taleck Exp $
 *
 * Copyright (c) 2003 Greg Taleck <taleck@oz.net>
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <pcap.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#ifdef HAVE_LIBTCPDUMP
#include "libtcpdump.h"
#endif


#include "dnet.h"
#include "misc.h"
#include "version.h"

#define URAND_MAX 0xffffffff

#undef FALSE
#define FALSE (0)
#undef TRUE
#define TRUE  (1)

#define DEBUG(fmt, args...) do {                \
    if (debug)                                  \
        fprintf(stderr, fmt, ##args);           \
} while (0)


#define RAND_IP6ADDR(addr) do {                              \
        ((u_int32_t *)(addr)->data)[0] = rand_uint32(rand);  \
        ((u_int32_t *)(addr)->data)[1] = rand_uint32(rand);  \
        ((u_int32_t *)(addr)->data)[2] = rand_uint32(rand);  \
        ((u_int32_t *)(addr)->data)[3] = rand_uint32(rand);  \
} while (0)

#define RAND_ETHMAC(addr) do {                              \
        ((u_int16_t *)(addr)->data)[0] = rand_uint16(rand);  \
        ((u_int16_t *)(addr)->data)[1] = rand_uint16(rand);  \
        ((u_int16_t *)(addr)->data)[2] = rand_uint16(rand);  \
} while (0)


#define MAX_PAYLOAD   (1514) /* XXX should get this via libdnet? */
#define DEFAULT_INTERFACE "fxp0"

static void usage();
u_short in_cksum_hdr(const struct ip_hdr *ip);
u_short in_cksum2(struct ip_hdr *ip, u_int8_t *data);
u_short in6_cksum(struct ip6_hdr *ip6, u_int8_t *data, u_int32_t proto);

int done = FALSE;


void
sig_handler(int signal)
{
    done = TRUE;
}

int putstderr(int ch)
{
    return fprintf(stderr, "%c", ch);
}

static int pkts_sent;
static int pkts_total;

void
do_progress(int flag)
{
    static int last_len = 0;
    char buf[1024];
    struct itimerval itv;

    bzero(buf, sizeof(buf));
    
    if (pkts_total == -1) { /* infinite */
        snprintf(buf, sizeof(buf), "\r { %10d pkts sent }", pkts_sent);
    } else {
        snprintf(buf, sizeof(buf), "\r { %10d/%d pkts sent %5.1f%% }", pkts_sent,
                 pkts_total, (float)(pkts_sent * 100) / (float)pkts_total);
    }
    
    write(fileno(stderr), buf, strlen(buf));
    fflush(stderr);
    last_len = strlen(buf);

    if (flag != 0) {
        signal(SIGALRM, do_progress);
        itv.it_value.tv_sec = 0;
        itv.it_value.tv_usec = 500000;
        itv.it_interval = itv.it_value;
        setitimer(ITIMER_REAL, &itv, NULL);
    } else {
        itv.it_value.tv_sec = 0;
        itv.it_value.tv_usec = 0;
        itv.it_interval = itv.it_value;
        setitimer(ITIMER_REAL, &itv, NULL);
        fprintf(stderr, "\n");
    }
}

struct ip6_exthdr {
    uint8_t  ext_nxt;       /* next header */
    uint8_t  ext_len;       /* following length in units of 8 octets */
};

int
main(int argc, char *argv[])
{
    int c, i, res;
    FILE *ttyout = stderr;
    struct pcap_pkthdr hdr;
    struct pcap_file_header file_hdr;
    eth_t *eth = NULL;
    rand_t *rand = NULL;
    int show_progress = TRUE;
    int do_tunnel;
    int attempts;
    int retries = 0;
    int proto;
    int payload_len;
    u_int64_t total_bytes = 0;
    u_int32_t *payload;
    u_char *end = NULL;
    int num_headers = 0;
    int dump_hex = 0;
    u_char *nxt;
    char *interface     = DEFAULT_INTERFACE;
    int write_to_intf   = TRUE;
    char *save_pcap     = NULL;
    int save_pcap_fd    = -1;
    struct ip_hdr *ip   = NULL;
    struct tcp_hdr *tcp = NULL;
    struct udp_hdr *udp = NULL;
    struct icmp_hdr *icmp = NULL;
    struct ip6_addr src_addr;
    struct ip6_addr dst_addr;
    struct eth_hdr *ehdr;
    char *src_str = NULL;
    char *dst_str = NULL;
    u_char pkt[4096 * 4];
    struct ip6_hdr *ip6 = NULL;
    int seed;
#ifdef HAVE_LIBTCPDUMP    
    int tcpdump = FALSE;
#endif    
    int debug = 0;
    int num_packets = -1;
    int max_payload = MAX_PAYLOAD;

    float frag_percent  = 0;
    float tcp_percent   = 0;
    float udp_percent   = 0;
    float icmp_percent  = 0;
    float v6tun_percent = 0;
    float v4tun_percent = 0;
    float tun_percent   = 0;
    float cksum_percent = 0;
    float value;
    
    int rand_dest = TRUE;
    int rand_src = TRUE;

    seed = getpid();

#ifdef HAVE_LIBTCPDUMP    
    Nflag++;
    nflag++;
    tflag++;
    Sflag++;
#endif    

    while ((c = getopt(argc, argv,
                       "tDHnhd:bc:e:f:i:p:r:s:vw:y:I:T:U:4:6:")) != EOF) {
        switch (c) {
        case 'b': /* no 'b'ar */
            show_progress = FALSE; 
            break;

        case '4':
            v4tun_percent = atof(optarg) / 100.0;
            break;

        case '6':
            v6tun_percent = atof(optarg) / 100.0;
            break;

        case 'c':
            cksum_percent = atof(optarg) / 100.0;
            break;
            
        case 'e':
            num_headers = atoi(optarg);
            break;

        case 'h':
            usage();
            /* NOTREACHED */

        case 'i':
            interface = optarg;
            break;

        case 'd':
            rand_dest = FALSE;
            dst_str = optarg;
            break;

        case 'n':
            write_to_intf = FALSE;
            break;
            
        case 's':
            rand_src = FALSE;
            src_str = optarg;
            break;

        case 'r':
            seed = atoi(optarg);
            break;

        case 'D':
            debug++;
            break;

        case 'H':
            dump_hex++;
            break;

        case 'p':
            num_packets = atoi(optarg);
            break;

        case 'f':
            frag_percent = atof(optarg) / 100.0;
            break;

        case 't':
#ifdef HAVE_LIBTCPDUMP      
            tcpdump = TRUE;
#else
            fprintf(stderr, "error: this program was built without HAVE_LIBTCPDUMP\n");
            fprintf(stderr, "get libtcpdump and rebuild if you want this.\n");
            usage();
#endif      
            break;

        /* protocol randomization */
        case 'T':
            tcp_percent = atof(optarg) / 100.0;
            break;

        case 'U':
            udp_percent = atof(optarg) / 100.0;
            break;

        case 'I':
            icmp_percent = atof(optarg) / 100.0;
            break;

        case 'y':
            max_payload = atoi(optarg);
            break;
            
        case 'v':
            fprintf(ttyout, "ip6sic %s (build %s)\n",
                    IP6SIC_VERSION, IP6SIC_DATE);
            exit(0);

        case 'w':
            save_pcap = optarg;
            break;

        default:
            fprintf(stderr, "error: unknown option %c\n\n", c);
            usage();
            /* NOTREACHED */
        }
    }

    fprintf(ttyout, "ip6sic v.%s (%s) %s\n", IP6SIC_VERSION, IP6SIC_DATE,
           IP6SIC_COPYRIGHT);

    /* setup some signal handling */
    signal(SIGTERM, &sig_handler);
    signal(SIGINT, &sig_handler);
    signal(SIGQUIT, &sig_handler);

    /* get the dest address to use */
    if ((dst_str && strncmp(dst_str, "rand", sizeof("rand") - 1) == 0)
        || rand_dest) {
        rand_dest = TRUE;
        fprintf(ttyout, " * randomizing destination ips\n");
    } else {
        fprintf(ttyout, " * looking up %s\n", dst_str);
        if (resolve_ip6(&dst_addr, dst_str) != 0) {
            fprintf(stderr, "error: bad dest: %s\n", dst_str);
            exit(-1);
        }
    }

    /* get the source address to use */
    if ((src_str && strncmp(src_str, "rand", sizeof("rand") - 1) == 0)
        || rand_src) {
        rand_src = TRUE;
        fprintf(ttyout, " * randomizing source ips\n");
    } else {
        fprintf(ttyout, " * looking up %s\n", src_str);
        if (resolve_ip6(&src_addr, src_str) != 0) {
            fprintf(stderr, "error: bad source: %s\n", src_str);
            exit(-1);
        }
    }

    /* check the payload length */
    if (max_payload > MAX_PAYLOAD)
        max_payload = MAX_PAYLOAD;
    fprintf(ttyout, " * maximum payload set to %d bytes\n", max_payload);

    /* check tunneling parameters */
    if ((v4tun_percent + v6tun_percent) > 1.0) {
        fprintf(ttyout, " * [clipping v6tun_percent]\n");
        v6tun_percent = 1.0 - v4tun_percent;
    }
    if (v4tun_percent > 0.0)
        fprintf(ttyout, " * IPv4 tunneling %.1f%%\n", v4tun_percent * 100.0);
    if (v6tun_percent > 0.0)
        fprintf(ttyout, " * IPv6 tunneling %.1f%%\n", v6tun_percent * 100.0);

    tun_percent = v4tun_percent + v6tun_percent;

    /* check protocol parameters */
    if (tcp_percent + udp_percent + icmp_percent > 1.0) {
        fprintf(stderr, "error: protocol percentages must "
                "not be greater than 100%%\n");
        exit(-1);
    }

    /* save to a pcap? */
    if (save_pcap != NULL) {
        if ((save_pcap_fd = open(save_pcap, O_CREAT|O_WRONLY,
                                 0000666)) < 0) {
            fprintf(stderr, "error: could not open %s for writing: %s\n",
                    save_pcap, strerror(errno));
            exit(-1);
        }
             
        file_hdr.magic = 0xa1b2c3d4;
        file_hdr.version_major = 2;
        file_hdr.version_minor = 4;
        file_hdr.thiszone = 0;
        file_hdr.snaplen = 1514;
        file_hdr.sigfigs = 0;
        file_hdr.linktype = DLT_EN10MB;

        fprintf(ttyout, " * writing packets to pcap file %s\n", save_pcap);

        if (write(save_pcap_fd, (char *)&file_hdr, sizeof(file_hdr)) < 0) {
            fprintf(stderr, "error: could not write to file %s\n", save_pcap);
            close(save_pcap_fd);
            save_pcap = NULL;
        }
    }
    
    /* do some initialization */
    if (write_to_intf) {
        if ((eth = eth_open(interface)) == NULL) {
            fprintf(stderr, "error: could not open interface %s\n",
                    interface);
            fprintf(stderr, "try using the -i option\n");
            usage();
        }
        fprintf(ttyout, " * writing packets to interface %s\n", interface);
    }
    
    if ((rand = rand_open()) == NULL) {
        fprintf(stderr, "error: could not open rand util\n");
        eth_close(eth);
        exit(-1);
    }

    /* seed the librand */
    fprintf(ttyout, " * using seed %d\n", seed);
    rand_set(rand, &seed, sizeof(seed));

    max_payload -= sizeof(struct eth_hdr) + sizeof(struct ip6_hdr);

    pkts_total = num_packets;
    if (show_progress)
        do_progress(1);
    
    /* loop and send some packets */
    for (i = 0; !done && (i < num_packets || num_packets == -1); i++) {
        tcp = NULL;
        udp = NULL;
        ip = NULL;
            
        /* decide if we'll tunnel this packet or not */
        do_tunnel = (rand_uint32(rand) <= (URAND_MAX * (tun_percent * 100)));
        DEBUG("do tunnel ? %d\n", do_tunnel);

        /* pick a protocol */
        value = (float)rand_uint32(rand) / (float)URAND_MAX;
        DEBUG("value %f\n", value);
        if (value <= tcp_percent) {
            /* set up a tcp header */
            proto = IPPROTO_TCP;
        } else if (value > tcp_percent
                   && value <= tcp_percent + udp_percent) {
            /* set up a udp header */
            proto = IPPROTO_UDP;
        } else if (value > tcp_percent + udp_percent && value <
                   icmp_percent + tcp_percent + udp_percent) {
            proto = IPPROTO_ICMPV6;
        } else {
            proto = rand_uint8(rand);
        }
        DEBUG("proto = %d (%s)\n", do_tunnel,
              proto == IPPROTO_TCP ? "TCP" :
              proto == IPPROTO_UDP ? "UDP" :
              proto == IPPROTO_ICMPV6 ? "ICMPv6" :
              proto == IPPROTO_ICMP ? "ICMP" : "??");

        payload_len = 14 + 60 + (do_tunnel ? 40 : 0)
            + (((double)rand_uint32(rand) / (double)0xffffffff)
               * (max_payload - (14 + 60 + (do_tunnel ? 40 : 0))));
        DEBUG("payload len = %d\n", payload_len);
        
        bzero(pkt, sizeof(pkt));
        ehdr = (struct eth_hdr *)pkt;

        RAND_ETHMAC(&ehdr->eth_src);
        RAND_ETHMAC(&ehdr->eth_dst);
        ehdr->eth_type = htons(ETH_TYPE_IPV6);

        ip6 = (struct ip6_hdr *)(ehdr + 1);

        ip6->ip6_hlim = (u_char)rand_uint32(rand);
        ip6->ip6_flow = rand_uint32(rand);
        ip6->ip6_plen = htons(payload_len - (sizeof(struct ip6_hdr)
                                             + sizeof(struct eth_hdr)));
        ip6->ip6_vfc  = 0x60;
        nxt = &ip6->ip6_nxt;
        end = (u_char *)(ip6 + 1);

        if (num_headers > 0) {
            /* tack on some IPv6 Extension headers */
            int cnt;
            for (cnt = 0; cnt < num_headers; cnt++) {
                struct ip6_exthdr *hdr = (struct ip6_exthdr *)end;
                hdr->ext_len = rand_uint8(rand);
                *nxt = rand_uint8(rand);
                nxt = &hdr->ext_nxt;
                end = (u_char *)(hdr + 1);
            }
        }
        
        if (rand_uint32(rand) <= (URAND_MAX * frag_percent)) {
            *nxt = IPPROTO_FRAGMENT;
        }

        if (do_tunnel) {
            /* one or the other */
            if (rand_uint32(rand) <= (URAND_MAX * (v4tun_percent /
                                       (v4tun_percent + v6tun_percent)))) {
                /* tunnel is v4 */
                ip = (struct ip_hdr *)(ip6 + 1);
                end = (u_char *)(ip + 1);

                ip6->ip6_nxt = IPPROTO_IPIP;
                ip->ip_v   = 4;
                ip->ip_hl  = 5;
                ip->ip_tos = rand_uint8(rand);
                ip->ip_len = htons(payload_len - 94);
                ip->ip_id  = rand_uint32(rand);
                ip->ip_off = 0;
                ip->ip_ttl = rand_uint8(rand);
                nxt = &ip->ip_p;
                ip->ip_sum = rand_uint16(rand);
                ip->ip_src = rand_uint32(rand);
                ip->ip_dst = rand_uint32(rand);
                ip->ip_sum = 0;
                ip->ip_p = proto;
                ip->ip_sum = in_cksum_hdr(ip);
                if (proto == IPPROTO_TCP)
                    tcp = (struct tcp_hdr *)(ip + 1);
                if (proto == IPPROTO_UDP) 
                    udp = (struct udp_hdr *)(ip + 1);
            } else {
                /* tunnel is v6 */
                struct ip6_hdr *new_ip6;
                ip6->ip6_nxt = IPPROTO_IPV6;
                new_ip6 = (struct ip6_hdr *)end;
                new_ip6->ip6_vfc = 0x60;
                new_ip6->ip6_hlim = rand_uint32(rand);
                new_ip6->ip6_flow = rand_uint32(rand);
                new_ip6->ip6_plen = htons(ntohs(ip6->ip6_plen)
                                          - sizeof(struct ip6_hdr));
                RAND_IP6ADDR(&new_ip6->ip6_dst);
                RAND_IP6ADDR(&new_ip6->ip6_dst);
                new_ip6->ip6_nxt = proto;
                nxt = &new_ip6->ip6_nxt;
                
                if (rand_dest)
                    RAND_IP6ADDR(&ip6->ip6_dst);
                else
                    memcpy(&ip6->ip6_dst, &dst_addr, sizeof(struct ip6_addr));
                
                if (rand_src)
                    RAND_IP6ADDR(&ip6->ip6_src);
                else
                    memcpy(&ip6->ip6_src, &src_addr, sizeof(struct ip6_addr));

                end = (u_char *)(new_ip6 + 1);
                if (proto == IPPROTO_TCP)
                    tcp = (struct tcp_hdr *)end;
                if (proto == IPPROTO_UDP)
                    udp = (struct udp_hdr *)end;
                ip6 = new_ip6;
            }
        } else {
            if (proto == IPPROTO_TCP)
                tcp = (struct tcp_hdr *)end;
            else if (proto == IPPROTO_UDP)
                udp = (struct udp_hdr *)end;
            else if (proto == IPPROTO_ICMP)
                icmp = (struct icmp_hdr *)end;
        }

        if (rand_dest)
            RAND_IP6ADDR(&ip6->ip6_dst);
        else
            memcpy(&ip6->ip6_dst, &dst_addr, sizeof(struct ip6_addr));

        if (rand_src)
            RAND_IP6ADDR(&ip6->ip6_src);
        else
            memcpy(&ip6->ip6_src, &src_addr, sizeof(struct ip6_addr));

        /* randomize the payload */
        for (payload = (uint32_t *)end;
             payload < (uint32_t *)(pkt + sizeof(pkt)); payload++)
            *payload = rand_uint32(rand);

        if (tcp) tcp->th_off = 5;
        *nxt = proto;
        
        /* cksum the payload after randomizing it */
        if (rand_uint32(rand) <= (URAND_MAX * cksum_percent)) {
            if (proto == IPPROTO_TCP && tcp) { /* do a TCP cksum */
                tcp->th_sum = 0;
                tcp->th_sum = (ip) ?
                    in_cksum2(ip, (u_int8_t *)tcp) :
                    in6_cksum(ip6, (u_int8_t *)tcp, IPPROTO_TCP);
            }

            if (proto == IPPROTO_UDP && udp) { /* do a UDP cksum */
                udp->uh_sum = 0;
                if (ip) {
                    udp->uh_ulen = ntohs(htons(ip->ip_len) - sizeof(struct ip_hdr));
                    udp->uh_sum = in_cksum2(ip, (u_int8_t *)udp);
                } else {
                    udp->uh_ulen = ip6->ip6_plen;
                    udp->uh_sum = in6_cksum(ip6, (u_int8_t *)udp, IPPROTO_UDP);
                }
            }
        }

        if (write_to_intf) {
            attempts = 0;
        retry:
            res = eth_send(eth, pkt, payload_len);
            if (errno == ENOBUFS) {
                if (attempts < 3) {
                    attempts++;
                    retries++;
                    usleep(10);
                    goto retry;
                }
            }
            if (errno == ENXIO) {
                fprintf(stderr, "try specifying the interface with -i\n");
                goto exit_now;
            }
        }

        if (dump_hex) 
            hexdump((const u_char *)pkt, payload_len);

        total_bytes += (u_int64_t)payload_len;

#ifdef HAVE_LIBTCPDUMP  
        if (tcpdump) {
            char buffer[1024];
            hdr.caplen = hdr.len = payload_len;
            libtcpdump_snprintf(buffer, sizeof(buffer), &hdr, pkt);
            printf("%s", buffer);
            fflush(stdout);
        }
#endif
        
        if (save_pcap) {
            struct timeval tp;
            gettimeofday(&tp, NULL);
            hdr.ts.tv_sec = tp.tv_sec;
            hdr.ts.tv_usec = tp.tv_usec;
            hdr.caplen = hdr.len = payload_len;
            
            if (write(save_pcap_fd, &hdr, sizeof(hdr)) < 0) {
                fprintf(stderr, "error: write: %s\n", strerror(errno));
                close(save_pcap_fd);
                save_pcap = NULL;
            }
            if (write(save_pcap_fd, pkt, payload_len) < 0) {
                fprintf(stderr, "error: write: %s\n", strerror(errno));
                close(save_pcap_fd);
                save_pcap = NULL;
            }
        }
        pkts_sent = i;
    }

    if (show_progress)
        do_progress(0);
    
 exit_now:
    fprintf(ttyout, " * %d packets sent (%ju bytes %juMB) (%d retries)\n", i, total_bytes,
           total_bytes / (1024 * 1024), retries);

    if (save_pcap) {
        fprintf(ttyout, " * closing pcap file %s\n", save_pcap);
        close(save_pcap_fd);
    }

    if (write_to_intf) {
        fprintf(ttyout, " * shutting down interface %s\n", interface);
        eth_close(eth);
    }
    
    fprintf(ttyout, " * exiting\n");
    return 0;
}


void
usage()
{
    fprintf(stderr,
            "usage: ip6sic [-DhHtv] [-i intf] [-d dst addr] [-s src addr]\n"
            "       [-e num hdrs] [-p num pkts] [-r seed] [-w pcap file]\n"
            "       [-y max payload] [-f fragment pct] [-T TCP pct] [-U UDP pct]\n"
            "       [-I ICMP pct] [-c checksum pct] [-4 IPv4 tunnel pct]\n"
            "       [-6 IPv6 tunnel pct]\n");
    exit(-1);
}
